#
# This file is part of pysimpledb.
#
# pysimpledb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pysimpledb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pysimpledb.  If not, see <http://www.gnu.org/licenses/>.
#
#      Author Daniel Mikusa <dan@mikusa.com>
#   Copyright Daniel Mikusa 2008
#

class AbstractDao(object):
    """
    A simple base class for creating data access objects.

    Each new DAO can simply inheit from this class and define its own set of
    queries.  A function is automatically generated for every query that is
    defined.

      Ex.
        class MyObjectDao(AbstractDao):
            def __init__(self, db):
                self.queries = {
                    'get': {
                        'sql': 'SELECT * FROM mytable WHERE id = :id',
                        'execType': 'queryForObject',
                        'resultType': MyObject
                    }
                }
                AbstractDao.__init__(self, db, queries)

        dao = MyObjectDao(db)
        obj = dao.get(1)

    As you can see from the example you define your queries inside of the
    __init__ function and then pass them to the super class's __init__ function.

    This will then automatically create a function for each query defined.
    """
    def __init__(self, db, queries):
        """
        Initializes an interal Sql object which is used to query the database.

        In addition, the following self.queries options are defined...

         defauktKeyParam  - the default column name used as the key of the map.

                execType  - the way that this message should be called.  This
                            corresponds to the functions that are defined by
                            the Sql object.  The following choices are
                            available...

                             queryForMap
                             queryForObject
                             queryForList
                             queryForScalar
                             insert
                             update
                             delete

                            See the Sql object for more information on how each
                            of these functions behaves.
        """
        self.sql = Sql(db, queries)
        for query in queries.values():
            if not query.has_key('execType'):
                if query.has_key('rowMapper'):
                    query['execType'] = query['rowMapper'].DEFAULT_EXEC_TYPE
                else:
                    query['execType'] = 'update'

    def __getattr__(self, attr):
        """
        Creates a dynamic function for each defined query.

        One dynamic function is created for each query and has the same name
        as the key in the queries dictionary.
        """
        if attr in self.sql.queries:
            val = self.sql.queries[attr]

            if val.has_key('execType'):
                execType = val['execType']
                execFunc = getattr(self.sql, execType)

                if execType in ('queryForMap',):
                    keyParam = None
                    if val.has_key('defaultKeyParam'):
                        keyParam = val['defaultKeyParam']
                    def callable(keyParam=keyParam, **kwargs):
                        return execFunc(attr, keyParam=keyParam, **kwargs)

                elif execType in ('queryForObject', 'queryForList'):
                    def callable(**kwargs):
                        return execFunc(attr, **kwargs)

                elif execType in ('update', 'batch'):
                    def callable(obj=None):
                        return execFunc(attr, obj)

                else:
                    raise TypeError("Invalid Query Type %s" % execType)

                return callable
            else:
                raise AttributeError("Missing method %s called." % attr)

class Sql(object):
    def __init__(self, db, queries):
        """
        Initializes database connection and available set of queries.

        The db object should be an database connection object returned by your
        db-api 2.0 driver's connect function.

        The queries data structure is a dictionary of queries that are defined.
        Each key in the dictionary is a query name and it's value is also a
        dictionary which configures the query.

        The following query options are available.

                sql - the sql query to use.  The query is a paramterized query
                      which should be done according to your database driver.
                      Look at the paramstyle option for your driver to determine
                      the exact syntax.

         parameterProvider

          rowMapper -
        
           insertId - this is required for insert operations.  It is the name
                      of the resultType object's member that will contain the
                      automatically generated insert id after an insert
                      statement executes.
        """
        for id, query in queries.iteritems():
            if not query.has_key('sql'):
                raise Exception("You Must Define an SQL Query! [%s]" % id)
        self.db = db
        self.queries = queries

    def _load_descriptions(self, cur):
        return [item[0] for item in cur.description]

    def queryForMap(self, queryId, keyParam, **kwargs):
        """
        Returns the results as a dictionary.

        Returns a dictionary that is keyed on the column specified in the
        key_param field.  The value of each dictionary element will be an
        object of the type specified by resultType.
        """
        cur = self.db.cursor()
        query = self.queries[queryId]
        data = {}
        try:
            # Map Parameters
            if 'parameterProvider' in query.keys():
                cur.execute(query['sql'],
                            query['parameterProvider'].provide())
            else:
                cur.execute(query['sql'], **kwargs)
            # Map Results
            descriptions = self._load_descriptions(cur)
            keyCol = descriptions.index(keyParam)
            for rownum, row in enumerate(cur.fetchall()):
                if row[keyCol] not in data.keys():
                    data[row[keyCol]] = []
                if 'rowMapper' in query.keys():
                    data[row[keyCol]].append(
                            query['rowMapper'].map(descriptions, row, rownum))
                else:
                    data[row[keyCol]].append(row)
        finally:
            cur.close()
        return data

    def queryForList(self, query_id, **kwargs):
        """
        Return the results of the query as a list.

        Returns the results of the query as a list of objects of type
        resultType.
        """
        cur = self.db.cursor()
        query = self.queries[query_id]
        data = []
        try:
            # Map Parameters
            if 'parameterProvider' in query.keys():
                cur.execute(query['sql'],
                            query['parameterProvider'].provide())
            else:
                cur.execute(query['sql'], **kwargs)
            # Map Results
            descriptions = self._load_descriptions(cur)
            for rownum, row in enumerate(cur.fetchall()):
                if 'rowMapper' in query.keys():
                    data.append(
                        query['rowMapper'].map(descriptions, row, rownum))
                else:
                    data.append(row)
        finally:
            cur.close()
        return data

    def queryForObject(self, query_id, **kwargs):
        """
        Returns a single object of type resultType.

        Returns a single object of type resultType.  The object will contain
        properties for each column in the query.  The property name will match
        that of the query exactly.

        TIP:  You can use a SELECT col AS XYZ to rename the parameters that
              will be set on the object.
        """
        cur = self.db.cursor()
        query = self.queries[query_id]
        data = None
        try:
            # Map Parameters
            if 'parameterProvider' in query.keys():
                cur.execute(query['sql'],
                            query['parameterProvider'].provide())
            else:
                cur.execute(query['sql'], **kwargs)
            # Map Results
            descriptions = self._load_descriptions(cur)
            row = cur.fetchone()
            if 'rowMapper' in query.keys():
                data = query['rowMapper'].map(descriptions, row, 1)
            else:
                data = row
        finally:
            cur.close()
        return data

    def batch(self, query_id, objs=[]):
        """Executes a batch of queries, not this will not update the id, autocommitted"""
        query = self.queries[query_id]
        # validate parameter type
        if (('paramClass' in query) and
            (not objs) and
            (len([obj for obj in objs if isinstance(obj, query['paramClass'])]))):
            raise Exception(
                'Object (%s) is not of correct type (%s) '
                    'specified by paramClass' %
                        (obj, query['paramClass']))
        # execute the statement
        cur = self.db.cursor()
        data = None
        try:
            if objs:
                cur.executemany(query['sql'], [obj.__dict__ for obj in objs])
            else:
                cur.executemany(query['sql'])
            self.db.commit()
            data = cur.rowcount
        finally:
            cur.close()
        return data

    def update(self, query_id, obj=None):
        """Executes a statement that changes the database, user must commit"""
        query = self.queries[query_id]
        # validate parameter type
        if (('paramClass' in query) and
            (not obj) and
            (not isinstance(obj, query['paramClass']))):
            raise Exception(
                'Object (%s) is not of correct type (%s) '
                    'specified by paramClass' %
                        (obj, query['paramClass']))
        # execute the statement
        cur = self.db.cursor()
        data = None
        try:
            if obj:
                cur.execute(query['sql'], obj.__dict__)
            else:
                cur.execute(query['sql'])
            if query.has_key('insertId'):
                obj.__dict__[query['insertId']] = cur.lastrowid
            data = cur.rowcount
        finally:
            cur.close()
        return data
    